#ifndef __bbemu_h__
#define __bbemu_h__

#include <stdio.h>
#include <netinet/in.h>

#include "parameters.h"
#include "bbqueue.h"

#define MAX_NODES 100
#define MAX_INSTANCES 100

#define panic(a...) do { fprintf(stderr, a); fprintf(stderr, "\n"); exit(1); } while (0)
#define dprintf(a...) do { } while (0)
#define log(a...) do { if (globalLogFile) { long long now = getTimeMicros(); fprintf(globalLogFile, "%d.%06d ", (int)(now/1000000ULL), (int)(now%1000000)); fprintf(globalLogFile, a); fprintf(globalLogFile, "\n"); if (globalLogFlush) fflush(globalLogFile); } } while (0)
#define event(a...) do { if (eventFile) { long long now = getTimeMicros(); fprintf(eventFile, "%d.%06d X", (int)(now/1000000ULL), (int)(now%1000000)); fprintf(eventFile, a); } } while (0)
//#define event(a...) do { } while (0)
#define event2(typ, data, size) do { if (eventFile && ((size)>=(16+28)) && ((data)[28+2] == 0xAF) && ((data)[28+3] == 0xFE)) { long long now = getTimeMicros(); fprintf(eventFile, "%d.%06d Y%s %d %08X %lld\n", (int)(now/1000000ULL), (int)(now%1000000), (typ), (data)[28+0], ntohl(*(int*)&(data)[28+4]), ntohll(*(long long*)&(data)[28+8])); } } while (0)

class Controller;
class Policy;

extern FILE *globalLogFile;
extern FILE *eventFile;
extern bool globalLogFlush;

void dump(FILE *stream, unsigned char *payload, int len);
unsigned short checksum(unsigned char *p1, int size1, unsigned char *p2, int size2);
void recalculateChecksum(unsigned char *payload, int len);
void adjustPacketIPs(unsigned char *data, int len, in_addr_t fromIP, in_addr_t toIP);
long long getTimeMicros();
unsigned long long ntohll(unsigned long long x);
unsigned long long htonll(unsigned long long x);

struct nodeInfo {
  in_addr_t externalAddr;
  int tunnelPort;
};

struct instanceInfo {
  Policy *policy;
  in_addr_t internalAddr;
  int portRangeFrom;
  int portRangeTo;
  int whichNode;
};

extern struct nodeInfo node[MAX_NODES];
extern struct instanceInfo instance[MAX_INSTANCES];
extern int numNodes;
extern int numInstances;
extern int localNode;

class Policy {
protected:
  Controller *myController;
  int myInstance;

public:
  Policy(Controller *controller, int instance) { this->myController = controller; this->myInstance = instance; };
  virtual ~Policy() {};
  virtual void upstreamPacket(int toInstance, unsigned char *data, int size) {};
  virtual void downstreamPacket(int fromInstance, unsigned char *data, int size) {};
  virtual void udpMessage(in_addr_t srcIP, int srcPort, unsigned char *data, int len) {};
  virtual void timerExpired(long long now) {};
};  

class ModelPolicy : public Policy {
protected:
  const static int MAX_PACKETS = 1000;  struct packetInfo {
    bool inUse;
    int from, to;
    int size;
    unsigned char buf[65536];
  } packet[MAX_PACKETS];

  int numSlotsUsed;
  BroadbandQueue *upQueue;
  BroadbandQueue *downQueue;
  long long currentTimer;

  int savePacket(int fromInstance, int toInstance, unsigned char *data, int size);
  void releasePacket(int id);
  void makeProgress(long long now);
  void updateTimer(void);

public:
  ModelPolicy(Controller *controller, int instance, ParameterFile *params);
  virtual void upstreamPacket(int toInstance, unsigned char *data, int size);
  virtual void downstreamPacket(int fromInstance, unsigned char *data, int size);
  virtual void timerExpired(long long now);
};

class PassThroughPolicy : public Policy {
public:
  PassThroughPolicy(Controller *controller, int instance, ParameterFile *params);
  virtual void upstreamPacket(int toInstance, unsigned char *data, int size);
  virtual void downstreamPacket(int fromInstance, unsigned char *data, int size);
};

class Controller {
public:
  Controller() {};
  virtual ~Controller() {};
  virtual void acceptUDPmessages(Policy *policy, in_addr_t srcip, int srcport, int dstport) {};
  virtual void acceptUDPmessagesNoMore(Policy *policy, in_addr_t srcip, int srcport, int dstport) {};
  virtual void sendUDPMessage(int localPort, in_addr_t destAddr, int destPort, unsigned char *data, int size) {};
  virtual void tunnelPacket(int fromInstance, int toInstance, unsigned char *data, int size) {};
  virtual void deliverPacket(int fromInstance, int toInstance, unsigned char *data, int size) {};
  virtual void setTimer(int instance, long long timeMicros) {};
  virtual long long lastWakeupMicros() { return 0; };
  virtual void run() {};
};

class LinuxController : public Controller {
protected:
  const static int MAX_LISTEN_PORTS = 20;
  const static int MAX_REGISTRATIONS = 20;
  struct udpRegistration {
    in_addr_t srcIP;
    int srcPort;
    int destPort;
    Policy *handler;
  } udpRegistration[MAX_REGISTRATIONS];
  struct listenPort {
    int port;
    int fd;
  } listenPort[MAX_LISTEN_PORTS];
  long long instanceTimer[MAX_INSTANCES];
    
  unsigned char tapMAC[6];
  long long nextTimer;
  long long lastWakeup;
  bool writeTimestamps;
  bool changeMTU;
  bool doPings;
  int receiveBufferSize;
  int numUDPRegistrations;
  int numListenPorts;
  int tunnelPort;
  int tunnelSock;
  int tunTapHandle;
  bool usingTAP;
  in_addr_t localIP;
  
  void forwardToDownstream(int fromInstance, int toInstance, unsigned char *data, int size);
  virtual int openTunTap(char *tapName, in_addr_t *tapIP, in_addr_t *tapMask, unsigned char *tapMAC);
  void addTimestamp(unsigned char *data, int size, bool hasIPheader, long long timestamp);
  int maybeModifyPacket(unsigned char *buffer, int size);
  
public:
  LinuxController(in_addr_t localIP, int port, ParameterFile *params);
  virtual void tunnelPacket(int fromInstance, int toInstance, unsigned char *data, int size);
  virtual void deliverPacket(int fromInstance, int toInstance, unsigned char *data, int size);
  virtual void acceptUDPmessages(Policy *policy, in_addr_t srcip, int srcport, int dstport);
  virtual void acceptUDPmessagesNoMore(Policy *policy, in_addr_t srcip, int srcport, int dstport);
  virtual void sendUDPMessage(int localPort, in_addr_t destAddr, int destPort, unsigned char *data, int size);
  virtual void setTimer(int instance, long long timeMicros);
  virtual long long lastWakeupMicros() { return lastWakeup; };
  virtual void run();
};

class PlanetlabController : public LinuxController {
protected:
  virtual int openTunTap(char *tapName, in_addr_t *tapIP, in_addr_t *tapMask, unsigned char *tapMAC);

public:
  PlanetlabController(in_addr_t localIP, int port, ParameterFile *params);
};

#endif /* defined(__bbemu_h__) */
