#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <linux/if_tun.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <unistd.h>
#include <assert.h>
#include <signal.h>

/* Controller for Planetlab
   This adapts the LinuxController for use with Planetlab hosts,
   which have some additional restrictions. */

#include "bbemu.h"

int PlanetlabController::openTunTap(char *tapName, in_addr_t *tapIP, in_addr_t *tapMask, unsigned char *tapMAC) 
{
  /* Open the TAP device. On PlanetLab, this exists permanently, 
     but working with it is practically an art. */

  int tapHandle;
  if ((tapHandle = open("/dev/net/tun", O_RDWR)) < 0) 
    panic("Cannot open /dev/net/tun (code %d). Is the tun module loaded and the device file available?", tapHandle);

  struct ifreq ifr;
  memset(&ifr, 0, sizeof(ifr));
  ifr.ifr_flags = IFF_TAP | IFF_NO_PI;

  if (ioctl(tapHandle, TUNSETIFF, (void*) &ifr) < 0)
    panic("Cannot set TAP device interface (%s)", strerror(errno));

  strcpy(tapName, ifr.ifr_name);

  /* Determine the TAP's IP and MAC addresses */

  int sd = socket(PF_INET, SOCK_STREAM, 0);
  if (sd < 0)
    panic("Cannot allocate socket (%s)", strerror(errno));

  if (ioctl(sd, SIOCGIFADDR, &ifr) == -1) 
    panic("Failed to get TAP IP addresss (%s)", strerror(errno));

  struct sockaddr_in addr;
  memcpy(&addr, &ifr.ifr_addr, sizeof(addr));
  *tapIP = addr.sin_addr.s_addr;

  if (ioctl(sd, SIOCGIFNETMASK, &ifr) == -1) 
    panic("Failed to get TAP netmask (%s)", strerror(errno));

  memcpy(&addr, &ifr.ifr_addr, sizeof(addr));
  *tapMask = addr.sin_addr.s_addr;
  
  if (ioctl(sd, SIOCGIFHWADDR, &ifr) == -1) 
    panic("Failed to get TAP MAC addresss (%s)", strerror(errno));

  memcpy(tapMAC, &ifr.ifr_hwaddr.sa_data, 6);  
  close(sd);

  usingTAP = true;
  return tapHandle;
}

PlanetlabController::PlanetlabController(in_addr_t localIP, int port, ParameterFile *params) : LinuxController(localIP, port, params)
{
}
