#include <assert.h>
#include <ctype.h>
#include <sys/time.h>

#include "bbemu.h"
#include "headers.h"

void dump(FILE *stream, unsigned char *payload, int len)
{
  int i, off = 0;
  while (off < len) {
    fprintf(stream, "%04X   ", off);
    for (i=0; i<16; i++) {
      if ((i+off) < len)
        fprintf(stream, "%02X ", payload[i+off]);
      else
        fprintf(stream, "   ");
    }
    
    fprintf(stream, "   ");
    
    for (i=0; i<16; i++) {
      if ((i+off) < len)
        fprintf(stream, "%c", isprint(payload[i+off]) ? payload[i+off] : '.');
    }
    
    off += 16;
    fprintf(stream, "\n");
  }
}

unsigned short checksum(unsigned char *p1, int size1, unsigned char *p2, int size2)
{
  assert(p1 && (size1>=0) && (size2>=0));
  unsigned int sum=0;

  for (int k=0; k<2; k++) {
    unsigned char *lpacket = (k==0) ? p1 : p2;
    if (lpacket) {
      int size = (k==0) ? size1 : size2;
      for (int i=0; i<size; i+=2) {
        unsigned short word16;
        if ((i+1)<size)
          word16 = (((unsigned int)lpacket[i])<<8)+lpacket[i+1];
        else
          word16 = (((unsigned int)lpacket[i])<<8);
        sum += (unsigned int) word16;
      }
    }
  }
    
  while (sum>>16)
    sum = (sum & 0xFFFF) + (sum>>16);
    
  return (~sum)&0xFFFF;
}

void recalculateChecksum(unsigned char *payload, int len)
{
  struct IpHeader *hip = (struct IpHeader*) payload;
  int ihl = (hip->ip_vhl&0xF)*4;
  assert((0<=ihl) && ((int)ihl<=len));
  hip->ip_xsum = 0;
  hip->ip_xsum = htons(checksum((unsigned char*) hip, 0, (unsigned char*) hip, ihl));
  
  /* TCP and UDP packets have a second checksum */
  
  if (hip->ip_proto == IPPROTO_TCP) {
    unsigned char *lpacket = (unsigned char*) hip;
    struct TcpHeader *htcp = (struct TcpHeader*)(&lpacket[ihl]);
    htcp->tcp_xsum = 0;
    struct tcpPseudoHdr pseudoHdr = { hip->ip_src, hip->ip_dst, 0, hip->ip_proto, htons(len - ihl) };
    htcp->tcp_xsum = htons(checksum((unsigned char*)&pseudoHdr, sizeof(struct tcpPseudoHdr), (unsigned char*)htcp, len-ihl));
  } else if (hip->ip_proto == IPPROTO_UDP) {
    unsigned char *lpacket = (unsigned char*) hip;
    struct UdpHeader *hudp = (struct UdpHeader*)(&lpacket[ihl]);
    hudp->checksum = 0;
    struct tcpPseudoHdr pseudoHdr = { hip->ip_src, hip->ip_dst, 0, hip->ip_proto, htons(len - ihl) };
    hudp->checksum = htons(checksum((unsigned char*)&pseudoHdr, sizeof(struct tcpPseudoHdr), (unsigned char*)hudp, len-ihl));
  }
}

void adjustPacketIPs(unsigned char *data, int len, in_addr_t fromIP, in_addr_t toIP)
{
  if (len < 20)
    return;
    
  *(int*)&data[12] = fromIP;
  *(int*)&data[16] = toIP;
  recalculateChecksum(data, len);
}

long long getTimeMicros()
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec * 1000000LL + tv.tv_usec;
}

unsigned long long ntohll(unsigned long long x)
{
  return htonl((x>>32)&0xFFFFFFFF) | (((long long)htonl(x&0xFFFFFFFF))<<32);
}

unsigned long long htonll(unsigned long long x)
{
  return htonl((x>>32)&0xFFFFFFFF) | (((long long)htonl(x&0xFFFFFFFF))<<32);
}

